% 'exoShape' is a Derived software by Stanley Durrleman, Copyright (C) INRIA (Asclepios team), All Rights Reserved, 2006-2009, version 1.0
%--------------------------------------------------------------------------
% Based on MATCHINE v1.0 software.
% Copyright Universit Paris Descartes
% Contributor: Joan Alexis GLAUNES (2006)
% alexis.glaunes@mi.parisdescartes.fr
% 
% This software is a computer program whose purpose is to calculate an optimal 
% diffeomorphic transformation in 3D-space that allows to match two datasets 
% like points, curves or surfaces.
% 
% This software is governed by the CeCILL-B license under French law and
% abiding by the rules of distribution of free software. You can use, 
% modify and/ or redistribute the software under the terms of the CeCILL-B 
% license as circulated by CEA, CNRS and INRIA at the following URL
% "http://www.cecill.info". 
% 
% As a counterpart to the access to the source code and rights to copy,
% modify and redistribute granted by the license, users are provided only
% with a limited warranty  and the software's author, the holder of the
% economic rights, and the successive licensors have only limited 
% liability. 
% 
% In this respect, the user's attention is drawn to the risks associated
% with loading, using, modifying and/or developing or reproducing the
% software by the user in light of its specific status of free software,
% that may mean that it is complicated to manipulate, and that also
% therefore means that it is reserved for developers and experienced
% professionals having in-depth computer knowledge. Users are therefore
% encouraged to load and test the software's suitability as regards their
% requirements in conditions enabling the security of their systems and/or 
% data to be ensured and, more generally, to use and operate it in the 
% same conditions as regards security. 
% 
% The fact that you are presently reading this means that you have had
% knowledge of the CeCILL-B license and that you accept its terms.
%--------------------------------------------------------------------------
 
% Any use of this code should make reference to:
% - S. Durrleman, X. Pennec, A. Trouve and N. Ayache, Statistical Models of Sets of Curves and Surfaces based on Currents, Medical Image Analysis, (2009), DOI: 10.1016/j.media.2009.07.007
% 
% AND to at least one of the following references:
% - [for the use of 'landmark or measure matching']:
% J. Glaunes, L. Younes and A. Trouve, Diffeomorphic matching of distributions: A new approach for unlabelled point-sets and sub-manifolds matching, Proc. of the 2004 IEEE Computer Society Conference on Computer Vision and Pattern Recognition. (CVPR'04), Vol. 2, pp. 712--718, DOI: 10.1109/CVPR.2004.81 
% 
% - [for the use of 'curve matching']:
% J. Glaunes, A. Qiu, M. Miller, L. Younes, Large Deformation Diffeomorphic Metric Curve Mapping, International Journal of Computer vision, Springer, 2008, Vol. 80, No. 3, pp. 317--336, DOI: 10.1007/s11263-008-0141-9 
% 
% - [for the use of 'surface matching']:
% M. Vaillant and J. Glaunes, Surface Matching via Currents, Proc. of Information Processing in Medical Imaging (IPMI'05), Lecture Notes in Computer Science vol. 3565, Springer 2005, pp. 381--392

function [x,J,energie,attache] = adaptdesc(x,f,g,options,varargin)

% Steepest descent with adaptive stepsize
%
% f functional
% g gradient of f
% x variable
% in structure options:
%   maxiter = maximum number of iterations; default 500;
%   stepsize = initial stepsize of descent; default 1/norm(initial gradient)
%   stepmult = increase ratio of stepsize after each minimization success;
%     default 1.2
%   stepdiv = decrease ratio of stepsize after each minimization failure;
%     default 2
%   breakratio = ratio for termination criteria; default 1e-4
%   loopbreak = maximum number of loops in each iteration
%   verbosemode if 1 then display evaluation of f at each step
%
% varargin additional arguments passed to f and g
%
% returns x variable at the end of minimization
% and J record of function evaluations
%


maxiter = 500;
[J(1) energie(1) attache(1)] = feval(f,x,varargin{:});
G = feval(g,x,varargin{:});
stepsize = J(1)/sum(G(:).^2);
stepmult = 1.2;
stepdiv = 2;
breakratio = 1e-4;
loopbreak = 40;
verbosemode = 0;

loadstruct('options');
                                                                                                                 
stepsize = J(1)/sum(G(:).^2);  % stepsize was overloaded by options structure, we reset it.

for iter = 1:maxiter
    G = feval(g,x,varargin{:});
    stepsize = stepsize * stepdiv;
    minimtest = 0;
    loop = 0;
    while(~minimtest & (loop<loopbreak)) 
        stepsize = stepsize / stepdiv;
        xnew = x - stepsize * G; 
        [J(iter+1) energie(iter+1) attache(iter+1)] = feval(f,xnew,varargin{:});
        minimtest = (J(iter+1) < J(iter));
        loop = loop + 1;
    end
    x = xnew;
    stepsize = stepsize * stepmult;
    if verbosemode
        disp(['iteration ',num2str(iter),...
            '     functional = ',num2str(J(iter)),...
            '     regularity = ',num2str(energie(iter)),...
            '     attache = ', num2str(attache(iter)),...
            '     stepsize = ', num2str(stepsize)]);
    end
    if dosaveiter
        feval(saveiter,x,J)
    end

    % termination criteria
    if (J(iter)-J(iter+1)) < breakratio*(J(1)-J(iter+1)) || loop==loopbreak
        return
    end
end

if verbosemode
    disp('maximum number of iterations exceeded');
end
